<?php

namespace Schaffen\Fiscal\FiscalDocument\Managers;

use Carbon\Carbon;
use Freelabois\LaravelQuickstart\Extendables\ManipulationManager;
use Illuminate\Support\Facades\DB;
use Schaffen\Core\Cep\Models\County;
use Schaffen\Core\Cep\Models\Uf;
use Schaffen\Core\Person\Models\Person;
use Schaffen\Fiscal\FiscalDocument\Models\TelecomServiceFiscalNote;
use Schaffen\Fiscal\FiscalDocument\Repositories\NfstiRepository;
use Schaffen\Fiscal\FiscalDocument\Repositories\NfstRepository;
use Schaffen\Fiscal\FiscalDocument\Managers\NfstiManager;
use Schaffen\Fiscal\FiscalDocument\Models\FiscalDocumentSerie;
use Schaffen\Fiscal\FiscalDocument\Models\TelecomServiceFiscalNoteFile;
use Schaffen\Fiscal\FiscalDocument\Models\TelecomServiceFiscalNoteItem;
use Storage;

class NfstManager extends ManipulationManager
{
     
    private $SERVIÇOS = "";
    private $CADASTRAIS = "";
    private $MASTER = "";
    private $ORDEM_LINHA_INIT = 1;
    private $ORDEM_LINHA_FIM = 1;
    private $MASTER_CODE = null;
    
    protected $validation = [
        self::CREATE => [],
        self::EDIT => []
    ];

    
    /**
     * @var NfstRepository
     */
    protected $repository;
    /**
     * @var NfstiRepository
     */
    protected $nfsti_repository;


    public function __construct(
        NfstRepository $repository,
        NfstiRepository $nfsti_repository
    )
    {
        $this->repository = $repository;
        $this->nfsti_repository = $nfsti_repository;
    }

   /**
     * @param $values
     * @param int|null $id
     * @param array $relations
     * @return TelecomServiceFiscalNote
     * @throws \Throwable
     */
    public function storeOrUpdate($values, int $id = null, $relations = []): TelecomServiceFiscalNote
    {
        return DB::transaction(function () use ($values, $id, $relations) {
            /** @var TelecomServiceFiscalNote $nfst */
            $nfst = parent::storeOrUpdate($values, $id, $relations);
            $nfsti = $values['servicos'];
            if($id != null){
                $serv = TelecomServiceFiscalNoteItem::where('nfsti_fk_nfst_codigo', $id)->delete();
            }

            foreach ($nfsti as $servico) {
                $servico['nfsti_fk_nfst_codigo'] = $nfst->nfst_codigo;
                $nfstiManager = app()->make(NfstiManager::class);
                $nfstiManager->storeOrUpdate($servico);
            }
            $nfst->load(['services']);
            return $nfst;
        });
    }

    public function destroy(int $id)
    {
        return DB::transaction(function () use ($id) {
            $nfsti = TelecomServiceFiscalNoteItem::where('nfsti_fk_nfst_codigo', $id)->delete();
            $nfst = parent::destroy($id);

            if($nfsti && $nfst){
                return "Nota removida com sucesso!";
            }
        });
    }

    public function setColumn($docString, $init, $final, $type, $value, $decimal = 2){
        $value = mb_convert_encoding($value, "ISO-8859-1", "UTF-8");
        $tam = $final + 1 - $init;
        
        if($type == 'N'){
            if(strpos($value, '.')){ 
                $value = number_format($value, $decimal, '', '');
            }
            $value = (String)$value;
            $value = preg_replace('/[^-?0-9\s]/', '', $value); // Removes special chars.
            
            $value = substr($value, 0, $tam);
            $tam_string = strlen((String)$value);
        }else if($type == 'X'){
            $modelos = array("/(á|à|ã|â|ä)/","/(Á|À|Ã|Â|Ä)/","/(é|è|ê|ë)/","/(É|È|Ê|Ë)/","/(í|ì|î|ï)/","/(Í|Ì|Î|Ï)/","/(ó|ò|õ|ô|ö)/","/(Ó|Ò|Õ|Ô|Ö)/","/(ú|ù|û|ü)/","/(Ú|Ù|Û|Ü)/","/(ñ)/","/(Ñ)/", "/(ç)/", "/(Ç)/");
            $aux = preg_replace($modelos,explode(" ","a A e E i I o O u U n N c C"),$value);
            $tam_string = strlen((String)$aux);

            $mapa = array("/( Da )/","/( De )/","/( Do )/","/( Das )/","/( Dos )/","/( E )/","/(D'a)/","/(D'o)/","/(D'e)/","/(D'á)/");
            $value = preg_replace($mapa,explode("."," da . de . do . das . dos . e .D'A.D'O.D'E.D'Á"), $value);
            $value = substr($value, 0, $tam);

            // $modelos = array("/(á|à|ã|â|ä)/","/(Á|À|Ã|Â|Ä)/","/(é|è|ê|ë)/","/(É|È|Ê|Ë)/","/(í|ì|î|ï)/","/(Í|Ì|Î|Ï)/","/(ó|ò|õ|ô|ö)/","/(Ó|Ò|Õ|Ô|Ö)/","/(ú|ù|û|ü)/","/(Ú|Ù|Û|Ü)/","/(ñ)/","/(Ñ)/", "/(ç)/", "/(Ç)/");
            // $value = preg_replace($modelos,explode(" ","a A e E i I o O u U n N c C"),$value);
            // $value = preg_replace('/([^A-Za-z0-9\s])/', ' ', $value); // Removes special chars.
            // $value = preg_replace('/\s+/', ' ',$value); // Removes multiple spaces .
            
            // $value = substr($value, 0, $tam);
            // $tam_string = strlen((String)$value);
        }
        
        if($tam > $tam_string){
            $mux = $tam - $tam_string;
            
            if($type == 'N'){ 
                $signal = str_contains($value, "-");
                if($signal){
                    $value = preg_replace('/[^0-9\s]/', '', $value);
                    $value = str_repeat(0, $mux) . (String)$value;
                    $value = '-' . $value;
                }else{
                    $value = str_repeat(0, $mux) . (String)$value;
                }
            }else if($type == 'X'){
                $value = (String)$value . str_repeat(' ', $mux);
            } 
        }
        
        $docString .= $value;
        return $docString;
    }

    public function generateKey($dados){
        // - CNPJ ou CPF do destinatário ou do tomador do serviço; 
        // - Número do documento fiscal; 
        // - Valor Total; 
        // - Base de Cálculo do ICMS; 
        // - Valor do ICMS. 
        // - Data de emissão; 
        // - CNPJ do emitente do documento fiscal

        $dado1 = $this->setColumn("", 1, 14, 'N', $dados->nfst_cpf_cnpj);
        $dado2 = $this->setColumn("", 1, 9, 'N', $dados->nfst_numero);
        $dado3 = $this->setColumn("", 1, 12, 'N', $dados->nfst_valor_total);
        $dado4 = $this->setColumn("", 1, 12, 'N', $dados->nfst_bc_icms);
        $dado5 = $this->setColumn("", 1, 12, 'N', $dados->nfst_icms_destacado);
        $dado6 = $this->setColumn("", 1, 8, 'N', $dados->nfst_data_emissao);
        $dado7 = $this->setColumn("", 1, 14, 'N', $dados->nfst_cnpj_emitente);
        $string = "{$dado1}{$dado2}{$dado3}{$dado4}{$dado5}{$dado6}{$dado7}";
        return md5($string);
    }

    public function getData($id){
        $this->ORDEM_LINHA_INIT = $this->ORDEM_LINHA_FIM;
        $dados = TelecomServiceFiscalNote::find($id);
        $sdf = FiscalDocumentSerie::where('sdf_fk_dof_codigo','NFST')->first();
        $dados->nfst_ref_item_nf = $this->ORDEM_LINHA_INIT;
        $dados->nfst_numero = $sdf->sdf_numero_inicial;
        $dados->nfst_codigo_autenticacao_digital = $this->generateKey($dados);
        $sdf->sdf_numero_inicial = $sdf->sdf_numero_inicial+1;
        $dados->save();
        $sdf->save();
       
        return $dados;
    }

    public function nomeCidade($uf, $cidade){
        $uf_cod = Uf::select('uf_codigo')->where('uf_sigla', strtoupper($uf))->first()->uf_codigo;
        $modelos = array("/(á|à|ã|â|ä)/","/(Á|À|Ã|Â|Ä)/","/(é|è|ê|ë)/","/(É|È|Ê|Ë)/","/(í|ì|î|ï)/","/(Í|Ì|Î|Ï)/","/(ó|ò|õ|ô|ö)/","/(Ó|Ò|Õ|Ô|Ö)/","/(ú|ù|û|ü)/","/(Ú|Ù|Û|Ü)/","/(ñ)/","/(Ñ)/", "/(ç)/", "/(Ç)/");
        $cidade = preg_replace($modelos,explode(" ","a A e E i I o O u U n N c C"),$cidade);
        $cidade = preg_replace('/([^A-Za-z0-9\s])/', ' ', $cidade); // Removes special chars.
        $codigo = County::where('mun_fk_uf_codigo', $uf_cod)->where('mun_nome', strtoupper($cidade))->first();
        if($codigo){
            return $codigo->mun_nome_ibge;
        }
        return '';
    }

    public function getIBGE($uf, $cidade){
        $uf_cod = Uf::select('uf_codigo')->where('uf_sigla', strtoupper($uf))->first()->uf_codigo;
        $modelos = array("/(á|à|ã|â|ä)/","/(Á|À|Ã|Â|Ä)/","/(é|è|ê|ë)/","/(É|È|Ê|Ë)/","/(í|ì|î|ï)/","/(Í|Ì|Î|Ï)/","/(ó|ò|õ|ô|ö)/","/(Ó|Ò|Õ|Ô|Ö)/","/(ú|ù|û|ü)/","/(Ú|Ù|Û|Ü)/","/(ñ)/","/(Ñ)/", "/(ç)/", "/(Ç)/");
        $cidade = preg_replace($modelos,explode(" ","a A e E i I o O u U n N c C"),$cidade);
        $cidade = preg_replace('/([^A-Za-z0-9\s])/', ' ', $cidade); // Removes special chars.

        $codigo = County::where('mun_fk_uf_codigo', $uf_cod)->where('mun_nome', strtoupper($cidade))->first();
        if($codigo){
            return $codigo->mun_codigo_ibge;
        }
        return 0;
        
    }

    

    public function setName($dados, $type){
        $files = TelecomServiceFiscalNoteFile::whereDate('nfsta_date_insert', Carbon::today())
        ->where('nfsta_tipo_arquivo', 1)->where('nfsta_modelo', $dados->nfst_modelo)->get();
        
        $person = Person::where('pes_identificador', $dados->nfst_cnpj_emitente)->first();
        $end = $person->address()->first();
   
        $UF = $this->setColumn("", 1, 2, 'X', $end->pee_uf);
        $CNPJ = $this->setColumn("", 1, 14, 'N', $dados->nfst_cnpj_emitente);
        $MM = $this->setColumn("", 1, 2, 'N', $dados->nfst_modelo);
        $SSS = $this->setColumn("", 1, 3, 'N', $dados->nfst_serie);
        $AAMM = $this->setColumn("", 1, 4, 'N', $dados->nfst_ano_mes_referencia);
        $ST = $this->setColumn("", 1, 1, 'X', $dados->nfst_situacao_documento);
        $T = $this->setColumn("", 1, 1, 'X', $type);
        $VVV = $this->setColumn("", 1, 3, 'N', count($files)+1);

        $string = "{$UF}{$CNPJ}{$MM}{$SSS}{$AAMM}{$ST}01{$T}.{$VVV}";
        return $string;
    }


    public function docMaster($dados){
        $linha = "";
        $linha = $this->setColumn($linha, 1, 14, 'N', $dados->nfst_cpf_cnpj);
        $linha = $this->setColumn($linha, 15, 28, 'X', $dados->nfst_ie);
        $linha = $this->setColumn($linha, 29, 63, 'X', $dados->nfst_razao_social);
        $linha = $this->setColumn($linha, 64, 65, 'X', $dados->nfst_end_uf);
        $linha = $this->setColumn($linha, 66, 66, 'N', $dados->nfst_classe_consumo);
        $linha = $this->setColumn($linha, 67, 67, 'N', $dados->nfst_fase_ou_tipo_utilizacao);
        $linha = $this->setColumn($linha, 68, 69, 'N', $dados->nfst_grupo_tensao);
        $linha = $this->setColumn($linha, 70, 81, 'X', $dados->nfst_cod_id_consumidor_assinante);
        $linha = $this->setColumn($linha, 82, 89, 'N', $dados->nfst_data_emissao);
        $linha = $this->setColumn($linha, 90, 91, 'N', $dados->nfst_modelo);
        $linha = $this->setColumn($linha, 92, 94, 'N', $dados->nfst_serie);
        $linha = $this->setColumn($linha, 95, 103, 'N', $dados->nfst_numero);
        $linha = $this->setColumn($linha, 104, 135, 'X', $dados->nfst_codigo_autenticacao_digital);
        $linha = $this->setColumn($linha, 136, 147, 'N', $dados->nfst_valor_total);
        $linha = $this->setColumn($linha, 148, 159, 'N', $dados->nfst_bc_icms);
        $linha = $this->setColumn($linha, 160, 171, 'N', $dados->nfst_icms_destacado);
        $linha = $this->setColumn($linha, 172, 183, 'N', $dados->nfst_op_isentas);
        $linha = $this->setColumn($linha, 184, 195, 'N', $dados->nfst_outros_valores);
        $linha = $this->setColumn($linha, 196, 196, 'X', $dados->nfst_situacao_documento);
        $linha = $this->setColumn($linha, 197, 200, 'N', $dados->nfst_ano_mes_referencia);
        $linha = $this->setColumn($linha, 201, 209, 'N', $dados->nfst_ref_item_nf);
        $linha = $this->setColumn($linha, 210, 221, 'X', $dados->nfst_nro_terminal_ou_uc);
        $linha = $this->setColumn($linha, 222, 222, 'N', $dados->nfst_tipo_pessoa);
        $linha = $this->setColumn($linha, 223, 224, 'N', $dados->nfst_tipo_cliente);
        $linha = $this->setColumn($linha, 225, 226, 'N', $dados->nfst_subclasse_consumo);
        $linha = $this->setColumn($linha, 227, 238, 'X', $dados->nfst_nro_terminal_principal);
        $linha = $this->setColumn($linha, 239, 252, 'N', $dados->nfst_cnpj_emitente);
        $linha = $this->setColumn($linha, 253, 272, 'X', $dados->nfst_nro_fat_comercial);
        $linha = $this->setColumn($linha, 273, 284, 'N', $dados->nfst_valor_total_fat_comercial);
        $linha = $this->setColumn($linha, 285, 292, 'N', $dados->nfst_data_leitura_anterior);
        $linha = $this->setColumn($linha, 293, 300, 'N', $dados->nfst_data_leitura_atual);
        $linha = $this->setColumn($linha, 301, 350, 'X', '');
        $linha = $this->setColumn($linha, 351, 358, 'N', '');
        $linha = $this->setColumn($linha, 359, 388, 'X', $dados->nfst_informacao_adicional);
        $linha = $this->setColumn($linha, 389, 393, 'X', '');
        $linha = $this->setColumn($linha, 394, 425, 'X', md5($linha));
        $this->MASTER .= $linha . "\r\n";        
    }

    public function docServices($dados){
        $arq = "";
        $dados->services = $dados->services->sortBy('nfsti_nro_ordem');
        foreach ($dados->services as $value) {
            $linha = "";
            $linha = $this->setColumn($linha, 1, 14, 'N', $dados->nfst_cpf_cnpj);
            $linha = $this->setColumn($linha, 15, 16, 'X', $dados->nfst_end_uf);
            $linha = $this->setColumn($linha, 17, 17, 'N', $dados->nfst_classe_consumo);
            $linha = $this->setColumn($linha, 18, 18, 'N', $dados->nfst_fase_ou_tipo_utilizacao);
            $linha = $this->setColumn($linha, 19, 20, 'N', $dados->nfst_grupo_tensao);
            $linha = $this->setColumn($linha, 21, 28, 'N', $dados->nfst_data_emissao);
            $linha = $this->setColumn($linha, 29, 30, 'N', $dados->nfst_modelo);
            $linha = $this->setColumn($linha, 31, 33, 'N', $dados->nfst_serie);
            $linha = $this->setColumn($linha, 34, 42, 'N', $dados->nfst_numero);
            $linha = $this->setColumn($linha, 43, 46, 'N', $value->nfsti_fk_cfop_codigo);
            $linha = $this->setColumn($linha, 47, 49, 'N', $value->nfsti_nro_ordem);
            $linha = $this->setColumn($linha, 50, 59, 'X', $value->nfsti_cod_referencia);
            $linha = $this->setColumn($linha, 60, 99, 'X', $value->nfsti_descricao);
            $linha = $this->setColumn($linha, 100, 103, 'N', $value->nfsti_classificacao);
            $linha = $this->setColumn($linha, 104, 109, 'X', $value->nfsti_unidade);
            $linha = $this->setColumn($linha, 110, 121, 'N', 0, 3);
            $linha = $this->setColumn($linha, 122, 133, 'N', 0, 3);
            $linha = $this->setColumn($linha, 134, 144, 'N', $value->nfsti_valor_total);
            $linha = $this->setColumn($linha, 145, 155, 'N', 0);
            $linha = $this->setColumn($linha, 156, 166, 'N', 0);
            $linha = $this->setColumn($linha, 167, 177, 'N', $value->nfsti_bc_icms);
            $linha = $this->setColumn($linha, 178, 188, 'N', $value->nfsti_valor_icms);
            $linha = $this->setColumn($linha, 189, 199, 'N', $value->nfsti_op_isentas);
            $linha = $this->setColumn($linha, 200, 210, 'N', $value->nfsti_outros_valores);
            $linha = $this->setColumn($linha, 211, 214, 'N', $value->nfsti_aliquota_icms);
            $linha = $this->setColumn($linha, 215, 215, 'X', $dados->nfst_situacao_documento);
            $linha = $this->setColumn($linha, 216, 219, 'X', $dados->nfst_ano_mes_referencia);
            $linha = $this->setColumn($linha, 220, 234, 'X', '');
            $linha = $this->setColumn($linha, 235, 246, 'N', $value->nfsti_quantidade, 3);
            $linha = $this->setColumn($linha, 247, 257, 'N', 0, 6);
            $linha = $this->setColumn($linha, 258, 263, 'N', $value->nfsti_aliquota_pis, 4);
            $linha = $this->setColumn($linha, 264, 274, 'N', $value->nfsti_valor_pis);
            $linha = $this->setColumn($linha, 275, 280, 'N', $value->nfsti_aliquota_cofins, 4);
            $linha = $this->setColumn($linha, 281, 291, 'N', $value->nfsti_valor_cofins);
            $linha = $this->setColumn($linha, 292, 292, 'X', '');
            $linha = $this->setColumn($linha, 293, 294, 'N', $dados->nfst_tipo_isencao);
            $linha = $this->setColumn($linha, 295, 299, 'X', '');
            $linha = $this->setColumn($linha, 300, 331, 'X', md5($linha));
            $arq .= $linha . "\r\n";
            $this->ORDEM_LINHA_FIM += 1;
        }
        $this->SERVIÇOS .= $arq;
    }

    public function docPersonalData($dados){
        $linha = "";
        $linha = $this->setColumn($linha, 1, 14, 'N', $dados->nfst_cpf_cnpj);
        $linha = $this->setColumn($linha, 15, 28, 'X', $dados->nfst_ie);
        $linha = $this->setColumn($linha, 29, 63, 'X', $dados->nfst_razao_social);
        $linha = $this->setColumn($linha, 64, 108, 'X', $dados->nfst_end_logradouro);
        $linha = $this->setColumn($linha, 109, 113, 'N', $dados->nfst_end_numero);
        $linha = $this->setColumn($linha, 114, 128, 'X', $dados->nfst_end_complemento);
        $linha = $this->setColumn($linha, 129, 136, 'N', $dados->nfst_end_cep);
        $linha = $this->setColumn($linha, 137, 151, 'X', $dados->nfst_end_bairro);
        $linha = $this->setColumn($linha, 152, 181, 'X', $this->nomeCidade($dados->nfst_end_uf, $dados->nfst_end_cidade));
        $linha = $this->setColumn($linha, 182, 183, 'X', $dados->nfst_end_uf);
        $linha = $this->setColumn($linha, 184, 195, 'X', $dados->nfst_telefone_contato);
        $linha = $this->setColumn($linha, 196, 207, 'X', $dados->nfst_cod_id_consumidor_assinante);
        $linha = $this->setColumn($linha, 208, 219, 'X', $dados->nfst_nro_terminal_ou_uc);
        $linha = $this->setColumn($linha, 220, 221, 'X', $dados->nfst_uf_terminal);
        $linha = $this->setColumn($linha, 222, 229, 'N', $dados->nfst_data_emissao);
        $linha = $this->setColumn($linha, 230, 231, 'N', $dados->nfst_modelo);
        $linha = $this->setColumn($linha, 232, 234, 'N', $dados->nfst_serie);
        $linha = $this->setColumn($linha, 235, 243, 'N', $dados->nfst_numero);
        $linha = $this->setColumn($linha, 244, 250, 'N', $this->getIBGE($dados->nfst_end_uf, $dados->nfst_end_cidade));
        $linha = $this->setColumn($linha, 251, 255, 'X', '');
        $linha = $this->setColumn($linha, 256, 287, 'X', md5($linha));
        $this->CADASTRAIS.= $linha . "\r\n";
    }

    public function createFile($name, $conteudo, $tipo, $modelo){
        $conteudo = mb_convert_encoding($conteudo, "UTF-8", "ISO-8859-1");
        $nfsta = new TelecomServiceFiscalNoteFile;
        $nfsta->nfsta_arquivo = $conteudo;
        $nfsta->nfsta_nome_arquivo = $name;
        $nfsta->nfsta_tipo_arquivo = $tipo;
        $nfsta->nfsta_modelo = $modelo;
        $nfsta->nfsta_usu_insert = 'ADMIN';
        $nfsta->nfsta_date_insert = now();
        $nfsta_save = $nfsta->save();
        if($tipo == 1){
            $this->MASTER_CODE = $nfsta->nfsta_codigo;
        }        
        return $nfsta_save;
    }

    
    public function geraDocumento($itens){
        if(count($itens) > 0){
            foreach ($itens as $id) {
                $dados = $this->getData($id);
                $this->docServices($dados);
                $this->docPersonalData($dados);
                $this->docMaster($dados);
            }
            $nameM = $this->setName($dados, 'M');
            $nameI = $this->setName($dados, 'I');
            $nameD = $this->setName($dados, 'D');

            $arqM = $this->createFile($nameM, $this->MASTER, 1, $dados->nfst_modelo);
            $arqI = $this->createFile($nameI, $this->SERVIÇOS, 2, $dados->nfst_modelo);
            $arqD = $this->createFile($nameD, $this->CADASTRAIS, 3, $dados->nfst_modelo);
            


            if($arqM && $arqI && $arqD){
                foreach ($itens as $id) {
                    $dados = TelecomServiceFiscalNote::find($id);
                    $dados->nfst_fk_nfsta_codigo = $this->MASTER_CODE;
                    $dados->save();
                }

                Storage::disk('doc_fiscal')->put($nameM, $this->MASTER);
                Storage::disk('doc_fiscal')->put($nameI, $this->SERVIÇOS);
                Storage::disk('doc_fiscal')->put($nameD, $this->CADASTRAIS);
                
                return "Documentos gerados com sucesso!";
            }else{
                return "Não foi possível gerar os documentos!";
            }
        }else{
            return "Nenhum item disponível para geração de arquivos!";
        }
    }

    


}